import os
import pytest

from flask import Flask
from flask_sqlalchemy import SQLAlchemy


INSTANCE_ROOT = ''

app = Flask(__name__)
app.config['SQLALCHEMY_DATABASE_URI'] = 'postgresql://lagring:lagring@localhost:5432/lagring_test'
app.config['ASSET_STORAGE_ROOT'] = os.path.join(os.getcwd(), 'test/var/assets')
app.config['ASSET_URL_ROOT'] = '/assets'

db = SQLAlchemy(app)

core = None
entity_base = None
model = None
instance = None
storage = None

IMAGE_PATH = os.path.join(INSTANCE_ROOT, 'test/data/image.jpg')
DIR_PATH = os.path.join(INSTANCE_ROOT, 'test/data')


def test_core():
    from lagring import LagringCore

    global core
    core = LagringCore(os.path.join(INSTANCE_ROOT, app.config['ASSET_STORAGE_ROOT']))


def test_model_definition():
    from lagring import Entity, Asset

    global entity_base
    entity_base = Entity
    entity_base.bind(core)

    class A(db.Model, entity_base):
        __tablename__ = 'test'

        id = db.Column(db.Integer(), primary_key=True)

        image = Asset()
        directory = Asset()

    global model
    model = A


def test_entity_type():
    from lagring import Entity, Asset

    class B(db.Model, Entity):
        __entitytype__ = 'special'
        __tablename__ = 'regular'

        id = db.Column(db.Integer(), primary_key=True)

        image = Asset()

    assert(B.entity_type == 'special')


def test_none_asset():
    with app.app_context():
        a = model(id=1)
        assert(not a.image)
        assert(not a.image.abs_path)

        global instance
        instance = a


def test_entity_id():
    assert(instance.entity_id == 1)


def test_write_to_model_field():
    with app.app_context():
        instance.image = IMAGE_PATH


def test_get_path_from_model():
    with app.app_context():
        assert(os.path.isfile(instance.image.abs_path))


def test_write_metadata():
    with app.app_context():
        instance.image = IMAGE_PATH, {'width': 100, 'height': 80}
        assert(instance.image.width == 100)
        assert(instance.image.height == 80)


def test_write_stream():
    with app.app_context():
        instance.image = open(IMAGE_PATH, 'rb')
        assert(os.path.isfile(instance.image.abs_path))


def test_delete_asset():
    from lagring.asset import NoneAssetInstance
    path = instance.image.abs_path
    del instance.image
    assert(isinstance(instance.image, NoneAssetInstance))


def test_write_directory():
    with app.app_context():
        instance.directory = DIR_PATH
        assert(os.path.isdir(instance.directory.abs_path))


def test_iterassets():
    with app.app_context():
        instance.image = IMAGE_PATH
        instance.directory = DIR_PATH

        n = 0
        assets = []

        for asset, name in instance.iterassets():
            n += 1
            assets.append((asset, name))
            assert(
                os.path.isfile(asset.abs_path)
                or
                os.path.isdir(asset.abs_path)
            )

        assert(assets[0][1] in ('image', 'directory'))
        assert(assets[1][1] in ('image', 'directory'))
        assert(assets[0][1] != assets[1][1])


def test_flask_lagring():
    from lagring import FlaskLagring

    global storage
    storage = FlaskLagring()
    storage.init_app(app)


def test_image_asset_def():
    from lagring.assets.image import ImageAsset

    class D(db.Model, storage.Entity):
        __tablename__ = 'test2'

        id = db.Column(db.Integer(), primary_key=True)

        image = ImageAsset(size=(100, 100))
        image_no_process = ImageAsset()

    global model
    model = D


def test_image_asset():
    with app.app_context():
        a = model(id=1)
        assert(not a.image)
        assert(not a.image.abs_path)

        global instance
        instance = a


def test_image_no_process():
    with app.app_context():
        instance.image_no_process = IMAGE_PATH
        assert(os.path.isfile(instance.image_no_process.abs_path))


def test_image_asset_write():
    with app.app_context():
        instance.image = IMAGE_PATH
        assert(os.path.isfile(instance.image.abs_path))
        assert(instance.image.width == 100)
        assert(instance.image.height == 100)


def test_clone_assets():
    a = model(id=1)
    b = model(id=2)

    a.image = IMAGE_PATH
    storage.clone_assets(a, b)
    assert(os.path.isfile(a.image.abs_path))
    assert(os.path.isfile(b.image.abs_path))


def test_directory_asset_def():
    from lagring.assets.directory import DirectoryAsset

    class E(db.Model, storage.Entity):
        __tablename__ = 'test3'

        id = db.Column(db.Integer(), primary_key=True)

        directory = DirectoryAsset()

    global model
    model = E


def test_directory_asset():
    with app.app_context():
        a = model(id=1)
        assert(not a.directory)
        assert(not a.directory.abs_path)

        global instance
        instance = a


def test_directory_asset_write_directory():
    with app.app_context():
        instance.directory = os.path.join(DIR_PATH, 'directory_asset')
        assert(os.path.isdir(instance.directory.abs_path))
        assert(os.path.isfile(os.path.join(instance.directory.abs_path, 'image.jpg')))


def test_directory_asset_write_zip():
    with app.app_context():
        b = model(id=2)
        b.directory = os.path.join(DIR_PATH, 'image.jpg.zip')
        assert(os.path.isdir(instance.directory.abs_path))
        assert(os.path.isfile(os.path.join(instance.directory.abs_path, 'image.jpg')))


def test_directory_asset_write_invalid_zip():
    from lagring import StorageException

    with pytest.raises(StorageException) as excinfo:
        with app.app_context():
            c = model(id=3)
            c.directory = os.path.join(DIR_PATH, 'image.jpg')

    assert 'Valid zip-archive expected' in str(excinfo.value)


def test_directory_asset_write_zip_from_stream():
    from werkzeug.datastructures import FileStorage
    with app.app_context():
        d = model(id=4)
        with open(os.path.join(DIR_PATH, 'image.jpg.zip'), 'rb') as stream:
            d.directory = FileStorage(stream, 'image.jpg.zip')
        assert (os.path.isdir(instance.directory.abs_path))
        assert (os.path.isfile(os.path.join(instance.directory.abs_path, 'image.jpg')))



import os
import zipfile
import shutil
from tempfile import gettempdir
from uuid import uuid4


from lagring import Asset, StorageException, AssetProcessingException


class DirectoryAsset(Asset):
    """
    Asset type to store directory assets. Source can be a directory or zip archive which
    is unpacked upon upload to the storage.
    """
    def _get_temp_path(self):
        return os.path.join(gettempdir(), str(uuid4()))

    def _unzip(self, path, storage, meta):
        if zipfile.is_zipfile(path):
            temp_dir = self._get_temp_path()

            def cleanup():
                shutil.rmtree(temp_dir)

            with zipfile.ZipFile(path, 'r') as z:
                z.extractall(temp_dir)
                temp_src = storage.asset_source_adapter(temp_dir)
                return temp_src, meta, cleanup
        else:
            raise AssetProcessingException('Valid zip-archive expected')

    def upload(self, storage, src, meta=None):
        if src.type == 'directory':
            return src, meta, None
        elif src.type == 'file':
            return self._unzip(src.path, storage, meta)
        elif src.type == 'stream':
            temp_path = self._get_temp_path()
            with open(temp_path, 'wb') as f:
                shutil.copyfileobj(src.stream, f)
            new_src, _, cleanup = self._unzip(temp_path, storage, meta)

            def cleanup2():
                cleanup()
                os.remove(temp_path)

            return new_src, meta, cleanup2
        else:
            raise StorageException('Unknown source type')



import os
from tempfile import gettempdir
from uuid import uuid4

from PIL import Image

from lagring import Asset, StorageException, AssetRequirementsException, AssetProcessingException
from lagring.logger import log


class ImageAsset(Asset):
    """
    Asset type to store image files

    Supported parameters:

    size - target size
    width - target width
    height - target height
        If asset size is set only by width or height,
        target size is calculated using image aspect ratio.
    transform - process method
        'crop' - resize and crop to size
        'fit' - fit to size, preserving the original aspect ratio
    size_constraint - size constraint value (see below)
    constraint_type - size constraint type:
        'none' - no constraint
        'min' - minimum size is set
        'max' - maximum size is set
        'exact' - exact target size is set
        If the size constrain is not met, the StorageException will be thrown.
        Process method set in "transform" will be also taken into account.
    lazy_init - specify callable to return asset parameters. Typical usecase --
        setup asset using parameters specified in the application config. Asset
        will be initialized on first access.
    """
    def __init__(self, init_lazy=None, **kwargs):
        if init_lazy:
            self.init_lazy = init_lazy
        else:
            self._init(**kwargs)

    def _init(self, size=None, width=None, height=None,
              transform='crop', size_constraint=None, constraint_type='none'):
        if transform not in ('crop', 'fit'):
            raise ValueError('Unknown transformation type')

        self.size = size
        self.width = width
        self.height = height

        if size:
            if height or width:
                raise StorageException('Use either size or width/height')
            else:
                self.width, self.height = self.size
        elif width and height:
            self.size = (width, height)

        self.transform = transform
        self.size_constraint = size_constraint
        self.constraint_type = constraint_type

        if constraint_type != 'none':
            if size_constraint:
                self.size_constraint = size_constraint
            else:
                self.size_constraint = size
            self.constraint_type = constraint_type
        else:
            if size_constraint:
                raise AssetRequirementsException('Size constraint is set but no constraint type is specified')

        self.init_lazy = None

    def __getattr__(self, item):
        if not self.init_lazy:
            # ленивая инициализация не задана, ничего не поделаешь
            raise AttributeError(item)
        else:
            self._init(**self.init_lazy())
            return getattr(self, item)

    def _target_size(self, original_size):
        if self.size:
            size = self.size
        else:
            width, height = original_size
            if self.width:
                height = int(height * (self.width / width))
                width = self.width
            elif self.height:
                width = int(width * (self.height / height))
                height = self.height
            size = width, height

        return size

    @staticmethod
    def _get_image_size(src):
        try:
            img = Image.open(src)
            return img.size
        except OSError:
            raise AssetProcessingException('Failed to open image')

    @staticmethod
    def _size_to_dict(size):
        '''
        (width, height) --> {'width': width, 'height': height}
        '''
        return {'width': size[0], 'height': size[1]}

    def _get_image_metadata(self, src):
        return self._size_to_dict(self._get_image_size(src))

    @staticmethod
    def _downsize_img(img, size, mode='crop'):
        if mode == 'crop':
            iw, ih = img.size
            width, height = size
            if iw > width or ih > height:
                iaspect = iw / ih
                taspect = width / height
                if iaspect > taspect:
                    left = int((iw - ih * taspect) / 2)
                    right = iw - left
                    crop_rect = (left, 0, right, ih)
                else:
                    upper = int((ih - iw / taspect) / 2)
                    lower = ih - upper
                    crop_rect = (0, upper, iw, lower)

                img = img.crop(crop_rect).resize(size, Image.BICUBIC)
        elif mode == 'fit':
            img.thumbnail(size, Image.BICUBIC)

        return img

    def _downsize(self, src, dest, size, mode='crop', force_name=False, force_format=None):
        """
        Crop and resize down to given size
        :param size: tuple (width, height)
        :param mode: crop - rezise and crop, fit - fit to size
        :return: modified image
        """
        try:
            img = Image.open(src)
        except OSError:
            raise AssetProcessingException('Failed to open image')

        format = img.format
        res_img = self._downsize_img(img, size, mode=mode)

        if force_format:
            new_format = force_format
        else:
            new_format = 'JPEG' if format == 'JPEG' else 'PNG'

        extension = new_format.lower()

        if force_name:
            new_path = dest
        else:
            new_path = dest + '.' + extension

        try:
            res_img.save(new_path, new_format)
        except OSError:
            raise AssetProcessingException('Failed to save temporary image')

        return new_path, extension

    @staticmethod
    def _get_temp_path(dir=None, ext=''):
        if ext and ext[:1] != '.':
            ext = '.' + ext
        return os.path.join(gettempdir() if dir is None else dir, str(uuid4()) + ext)

    def upload(self, storage, src, meta=None):
        original_size = self._get_image_size(src.stream)

        if self.constraint_type == 'min':
            if original_size[0] < self.size_constraint[0] or original_size[1] < self.size_constraint[1]:
                raise StorageException('Size is less than minimum constraint')
        elif self.constraint_type == 'max':
            if original_size[0] > self.size_constraint[0] or original_size[1] > self.size_constraint[1]:
                raise StorageException('Size is bigger than maxmimum constraint')
        elif self.constraint_type == 'exact':
            if original_size[0] != self.size_constraint[0] and original_size[1] != self.size_constraint[1]:
                raise StorageException('Exact size constraint is not met')
        
        if not any([self.size, self.width, self.height]):
            # без обработки
            new_path = self._get_temp_path()
            if meta is None:
                meta = self._size_to_dict(original_size)

            try:
                img = Image.open(src.stream)
            except OSError:
                raise AssetProcessingException('Failed to open image')

            new_format = 'JPEG' if img.format == 'JPEG' else 'PNG'

            try:
                img.save(new_path, new_format)
            except OSError:
                raise AssetProcessingException('Failed to save temporary image')

            extension = new_format.lower()
        else:
            target_size = self._target_size(original_size)
            temp = self._get_temp_path()
            new_path, extension = self._downsize(src.stream, temp, target_size, self.transform)
            meta = self._get_image_metadata(new_path)
            log.debug(
                'Processing image: {} {}'.format(self.transform, target_size))

        new_src = storage.asset_source_adapter(new_path, extension=extension)

        def when_done():
            os.remove(new_path)

        return new_src, meta, when_done



import os

from werkzeug.datastructures import FileStorage

from .core import LagringCore, _AssetSource
from .entity import Entity
from .classproperty import classproperty
from .exception import StorageException


class _AssetSourceEx(_AssetSource):
    @property
    def type(self):
        if isinstance(self.src, FileStorage):
            if self.src.filename != '':
                return 'stream'
            else:
                return None
        else:
            return super().type

    @property
    def stream(self):
        if isinstance(self.src, FileStorage):
            if self.src.filename != '':
                return self.src.stream
            else:
                return None
        else:
            return super().stream


class FlaskLagring(LagringCore):
    def __init__(self):
        pass

    def init_app(self, app):
        root = app.config['ASSET_STORAGE_ROOT']
        if root[:1] != '/':
            raise StorageException('ASSET_STORAGE_ROOT path must be absolute')
        self.setup(
            root=root,
            url_base=app.config['ASSET_URL_ROOT'],
            write_allowed=True
        )

    def setup(self, root, url_base='', write_allowed=True):
        '''
        root            - file storage base directory
        url_base        - URL base
        write_allowed   - set this to False for read-only access
        '''
        LagringCore.__init__(self, root, url_base, write_allowed)

    @property
    def Entity(self):
        '''
        Asset-aware entity mixin class
        '''
        base = Entity
        base.bind(self)
        return base

    @classproperty
    def asset_source_adapter(cls):
        return _AssetSourceEx

    def clone_assets(self, src_entity, dest_entity):
        for asset, name in src_entity.iterassets():
            self.add(
                src_entity.entity_type, 
                dest_entity.id, 
                name, 
                self.asset_source_adapter(asset.abs_path)
            )
            data = src_entity.get_asset_data(name)
            dest_entity.put_asset_data(name, data)



# coding=utf-8

import os
import io
import shutil
import uuid

from .classproperty import classproperty
from .exception import StorageException
from .logger import log


class _AssetSource:
    def __init__(self, src, extension=None):
        self.src = src
        self.extension = self._normalise_extension(extension)

    @staticmethod
    def _normalise_extension(extension):
        if extension:
            return extension if extension[:1] == '.' else '.' + extension
        else:
            return ''

    def __bool__(self):
        return self.type is not None

    @property
    def type(self):
        if self.src is None:
            return None
        elif isinstance(self.src, str):
            if os.path.isfile(self.src):
                return 'file'
            elif os.path.isdir(self.src):
                return 'directory'
            else:
                raise StorageException('Path not found')
        elif isinstance(self.src, io.IOBase):
            return 'stream'
        else:
            raise StorageException('Unknown source type')

    @property
    def stream(self):
        if self.src is None:
            return None
        elif isinstance(self.src, str):
            if os.path.isfile(self.src):
                return open(self.src, 'rb')
            else:
                raise StorageException('File not found')
        elif isinstance(self.src, io.IOBase):
            return self.src
        else:
            raise StorageException('Unknown source type')

    @property
    def path(self):
        if isinstance(self.src, str):
            if os.path.isfile(self.src) or os.path.isdir(self.src):
                return self.src
        else:
            return None


class LagringCore(object):
    '''
    Storage base functionality
    '''

    def __init__(self, root, url_base='', write_allowed=True):
        '''
        root            - file storage base directory
        url_base        - URL base
        write_allowed   - set this to False for read-only access
        '''

        self.url_base = url_base
        self.root = root
        self.write_allowed = write_allowed

        if not os.path.exists(self.root):
            os.makedirs(self.root)
            log.info('Storage root has been created in "{}"'.format(self.root))

    @classproperty
    def asset_source_adapter(cls):
        return _AssetSource

    def _put(self, src, bucket, filename):
        '''
        Save asset file
        src - source
        bucket - storage directory
        filename - filename without directory
        '''
        bucket_path = os.path.join(self.root, bucket)
        if not os.path.exists(bucket_path):
            os.makedirs(bucket_path)

        dest_path = os.path.join(bucket_path, filename)

        type = src.type
        if type == 'file':
            shutil.copy(src.path, dest_path)
        elif type == 'directory':
            shutil.copytree(src.path, dest_path)
        elif type == 'stream':
            with open(dest_path, 'wb') as dest_obj:
                shutil.copyfileobj(src.stream, dest_obj)

        asset_path = os.path.join(bucket, filename)
        log.debug('Added "{}"'.format(asset_path))

        return asset_path

    def _trash(self, asset_path):
        '''
        Delete file
        '''
        abs_path = self.abs_path(asset_path)
        if os.path.isfile(abs_path):
            os.remove(abs_path)
        elif os.path.isdir(abs_path):
            shutil.rmtree(abs_path)
        else:
            log.warning('Non-existent path to delete "{}"'.format(asset_path))
            return
        log.debug('Deleted "{}"'.format(asset_path))

    def _bucket(self, entity_type, entity_id, asset_name):
        '''
        Storage bucket path
        '''
        hex = '{0:0{1}x}'.format(entity_id, 8)
        return '{}/{}/{}/{}/{}'.format(entity_type, asset_name, hex[0:2], hex[2:4], hex[4:6])

    def add(self, entity_type, entity_id, asset_name, src):
        '''
        Add file or directory to the storage
        New filename contains random part so the old version must be explicitly deleted
        Returns path to the asset
        '''

        if not self.write_allowed:
            raise StorageException("Storage is in read-only mode")

        # генерим путь к ассету
        bucket = self._bucket(entity_type, entity_id, asset_name)
        filename = '{}-{}{}'.format(
            entity_id,
            str(uuid.uuid4())[:6],  # some random
            src.extension)

        asset_path = self._put(src, bucket, filename)

        log.debug(
            'Asset added: entity=({}, {}), asset_name={}'\
                .format(entity_type, entity_id, asset_name))

        return asset_path

    def delete(self, path):
        '''
        Delete asset
        '''

        if not self.write_allowed:
            raise StorageException("Storage is in read-only mode")

        self._trash(path)

    def url(self, path):
        '''
        Get asset URL
        '''

        return self.url_base + \
                   ('' if self.url_base.endswith('/') else '/') + \
                   path

    def abs_path(self, path):
        '''
        Get asset absolute path
        '''
        return os.path.join(self.root, path)



from .core import LagringCore
from .entity import Entity
from .asset import Asset, AssetInstance, NoneAssetInstance
from .exception import *
from .logger import log
from .flask_lagring import FlaskLagring



from .exception import StorageException


class AssetInstance:
    def __init__(self, url, path, abs_path, meta):
        self.url = url
        self.path = path
        self.abs_path = abs_path
        self.meta = meta

    def __bool__(self):
        return True

    def __getattr__(self, name):
        if name in self.meta:
            return self.meta[name]


class NoneAssetInstance:
    def __init__(self):
        self.url = None
        self.path = None
        self.abs_path = None
        self.meta = None

    def __bool__(self):
        return False


class Asset:
    def __get__(self, entity, entity_cls=None):
        if entity:
            # access from entity instance - return asset instance object
            return self._asset_instance(entity)
        else:
            # access from entity class - return asset definition object
            return self

    def __set__(self, entity, value):
        # split value to asset source and metadata dictionary
        if isinstance(value, tuple):
            if len(value) != 2:
                raise ValueError('Tuple of two elements expected')
            src, meta = value
        else:
            src, meta = value, None

        storage = entity.storage

        asset_source = storage.asset_source_adapter(src)
        if not asset_source:
            # empty asset sources are ignored
            return
        asset_source, meta, when_done = self.upload(storage, asset_source, meta)

        # delete existing asset if any
        asset = self._asset_instance(entity)
        if asset:
            storage.delete(asset.path)

        entity_type = entity.entity_type
        entity_id = entity.entity_id
        if not entity_id:
            raise StorageException('Entity does not have an id. Probably you forgot to flush before asset assignment.')
        asset_name = entity._asset_name(self)

        data = {
            'path': storage.add(entity_type, entity_id, asset_name, asset_source)
        }
        # append metadata fields
        if meta:
            for k, v in meta.items():
                data[k] = v

        # put to database
        entity.put_asset_data(asset_name, data)

        if when_done:
            when_done()

    def __delete__(self, entity):
        asset = self._asset_instance(entity)
        if asset:
            asset_name = entity._asset_name(self)
            entity.storage.delete(asset.path)
            entity.put_asset_data(asset_name, {})

    def _asset_instance(self, entity):
        """
        Build asset instance object from entity
        """
        name = entity._asset_name(self)
        value = entity.get_asset_data(name)
        if value:
            path = value['path']
            asset = AssetInstance(
                path=path,
                url=entity.storage.url(path),
                abs_path=entity.storage.abs_path(path),
                meta=value
            )
            return asset
        else:
            return NoneAssetInstance()

    def upload(self, storage, src, meta=None):
        """
        Preprocess file before actual loading to storage.
        Returns path to new file, altered/created metadata and when_done callable
        to be called when everything is done (to perform some cleanup if needed)
        
        """
        return src, meta, None

    def after_upload(self, storage, asset):
        """
        Post-process file after upload to storage
        """
        pass


from setuptools import setup, find_packages

setup(
    name="Lagring",
    version="0.2.6.2",
    author="Lenar Imamutdinov",
    author_email="lenar.imamutdinov@gmail.com",
    packages=find_packages(),
    include_package_data=True,
    url="https://github.com/neoden/lagring",
    license="MIT",
    description="Asset storage for Flask",
    long_description=open("README").read(),
    install_requires=[
        "flask",
        "sqlalchemy"
    ],
    extras_require = {
        'Pillow': []
    }
)


from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm.attributes import flag_modified
from sqlalchemy import Column

from .classproperty import classproperty
from .exception import StorageException
from .asset import Asset


class Entity:
    """
    Base class for asset-aware models
    """

    @declared_attr
    def _assets(self):
        return Column(JSONB)

    @classmethod
    def _build_asset_name_table(cls):
        if not hasattr(cls, '__asset_names'):
            fields = {}
            for name in dir(cls):
                if not name.startswith('_'):
                    asset_field = getattr(cls, name)
                    if isinstance(asset_field, Asset):
                        fields[asset_field] = name
            cls.__asset_names = fields

    @classproperty
    def _asset_names(cls):
        try:
            return cls.__asset_names
        except AttributeError:
            cls._build_asset_name_table()
            return cls.__asset_names

    @classmethod
    def _asset_name(cls, asset_def_obj):
        """
        Find asset name by asset definition object
        """
        try:
            return cls._asset_names[asset_def_obj]
        except KeyError:
            raise StorageException('Asset name not found')

    @classmethod
    def bind(cls, storage):
        """
        Late binding to storage instance
        """
        cls.__storage = storage

    @classproperty
    def storage(cls):
        """
        Storage instance
        """
        return cls.__storage

    @classproperty
    def entity_type(cls):
        """
        Unique entity type name. Specified by __entitytype__ or __tablename__.
        """
        if hasattr(cls, '__entitytype__'):
            return getattr(cls, '__entitytype__')
        elif hasattr(cls, '__tablename__'):
            return getattr(cls, '__tablename__')
        else:
            raise StorageException('Unknown entity type. Neither __entitytype__ nor __tablename__ are defined in the model.')

    @property
    def entity_id(self):
        """
        Return model id field. It is highly recommended to have an integer id field.
        """
        try:
            if self.id:
                return self.id
            raise StorageException('No entity id')
        except AttributeError:
            raise StorageException('No "id" column found. You might wish to override entity_id to handle non-trivial primary key fields.')

    def iterassets(self):
        """
        Iterate of entity instance assets
        """
        return ((getattr(self, name), name) for asset, name in self._asset_names.items())

    def asset_data_field(self, asset_name):
        """
        Return model field name that contains asset data for that asset name.
        Everything is stored in _assets field by default
        """
        return '_assets'

    def get_asset_data(self, name):
        """
        Extract given asset data from the model field
        """
        value = getattr(self, self.asset_data_field(name)) or {}
        return value.get(name)

    def put_asset_data(self, name, data):
        """
        Write asset data to the model field
        """
        field_name = self.asset_data_field(name)
        value = getattr(self, field_name) or {}
        value[name] = data
        setattr(self, field_name, dict(value))
        # without that JSON fields are not updated in commit
        flag_modified(self, field_name)



# coding=utf-8

class StorageException(Exception):
    """
    Storage-specific exception
    """
    pass



class AssetRequirementsException(StorageException):
    """
    Requirements for the asset data not met
    """
    pass


class AssetProcessingException(StorageException):
    """
    Error when processing asset data
    """
    pass
